
package DE.fub.inf.JVM.JavaClass;


import  DE.fub.inf.JVM.Constants;
import  java.io.*;

/**
 * This class represents an entry in the exception table of the <em>Code</em>
 * attribute and is used only there. It contains a range in which a
 * particular exception handler is active.
 *
 * @version 970922
 * @author  <A HREF="http://www.inf.fu-berlin.de/~dahm">M. Dahm</A>
 * @see     Code
 */
public final class CodeException
{
  private short start_pc;   // Range in the code the exception handler is
  private short end_pc;     // active. start_pc is inclusive, end_pc exclusive
  private short handler_pc; /* Starting address of exception handler, i.e.
			     * an offset from start of code.
			     */
  private short catch_type; /* If this is zero the handler catches any
			     * exception, otherwise it points to the
			     * exception class which is to be caught.
			     */


  /**
   * Empty constructor, all attributes have to be defined via `setXXX'
   * methods. Use at your own risk.
   */
  public CodeException() {}    
  /**
   * Initialize from another object.
   */
  public CodeException(CodeException c) {
    this(c.getStartPC(), c.getEndPC(), c.getHandlerPC(), c.getCatchType());
  }    
  /**
   * Construct object from file stream.
   * @param file Input stream
   * @throw IOException
   */  
  CodeException(DataInputStream file) throws IOException
  {
    this((short)file.readUnsignedShort(), (short)file.readUnsignedShort(),
	 (short)file.readUnsignedShort(), (short)file.readUnsignedShort());
  }    
  /**
   * @param start_pc Range in the code the exception handler is active,
   * start_pc is inclusive while
   * @param end_pc is exclusive
   * @param handler_pc Starting address of exception handler, i.e.
   * an offset from start of code.
   * @param catch_type If zero the handler catches any 
   * exception, otherwise it points to the exception class which is 
   * to be caught.
   */
  public CodeException(short start_pc, short end_pc, short handler_pc,
		       short catch_type)
  {
    this.start_pc   = start_pc;
    this.end_pc     = end_pc;
    this.handler_pc = handler_pc;
    this.catch_type = catch_type;
  }    
  /**
   * Called by objects that are traversing the nodes of the tree implicitely
   * defined by the contents of a Java class. I.e., the hierarchy of methods,
   * fields, attributes, etc. spawns a tree of objects.
   *
   * @param v Visitor object
   */
  public void accept(Visitor v) {
    v.visitCodeException(this);
  }    
  /**
   * Dump code exception to file stream in binary format.
   *
   * @param file Output file stream
   * @throw IOException
   */ 
  public final void dump(DataOutputStream file) throws IOException
  {
    file.writeShort(start_pc);
    file.writeShort(end_pc);
    file.writeShort(handler_pc);
    file.writeShort(catch_type);
  }    
  /**
   * @return Zero, if the handler catches any exception, otherwise it points to
   * the exception class which is to be caught.
   */  
  public final short getCatchType() { return catch_type; }    
  /**
   * @return Exclusive end index of the region where the handler is active.
   */  
  public final short getEndPC() { return end_pc; }    
  /**
   * @return Starting address of exception handler, relative to the code.
   */  
  public final short getHandlerPC() { return handler_pc; }    
  /**
   * @return Inclusive start index of the region where the handler is active.
   */  
  public final short getStartPC() { return start_pc; }    
  /**
   * @param catch_type.
   */
  public final void setCatchType(short catch_type) {
    this.catch_type = catch_type;
  }    
  /**
   * @param end_pc.
   */
  public final void setEndPC(short end_pc) {
    this.end_pc = end_pc;
  }    
  /**
   * @param handler_pc.
   */
  public final void setHandlerPC(short handler_pc) {
    this.handler_pc = handler_pc;
  }    
  /**
   * @param start_pc.
   */
  public final void setStartPC(short start_pc) {
    this.start_pc = start_pc;
  }    
  /**
   * @return String representation.
   */ 
  public final String toString()
  {
    return "CodeException(start_pc = " + start_pc + 
      ", end_pc = " + end_pc +
      ", handler_pc = " + handler_pc + ", catch_type = " + catch_type + ")";
  }    
}
