
package DE.fub.inf.JVM.JavaClass;


import  DE.fub.inf.JVM.Constants;
import  java.io.*;

/** 
 * This class represents a inner class attribute, i.e. the class
 * indices of the inner and outer classes, the name and the attributes
 * of the inner class.
 *
 * @version 970922
 * @author  <A HREF="http://www.inf.fu-berlin.de/~dahm">M. Dahm</A>
 * @see InnerClasses
 */
public final class InnerClass implements Constants
{
  private short inner_class_index;
  private short outer_class_index;
  private short inner_name_index;
  private short inner_access_flags;
  


  /**
   * Empty constructor, all attributes have to be defined via `setXXX'
   * methods. Use at your own risk.
   */
  public InnerClass() {}    
  /**
   * Initialize from another object.
   */
  public InnerClass(InnerClass c) {
    this(c.getInnerClassIndex(), c.getOuterClassIndex(), c.getInnerNameIndex(),
	 c.getInnerAccessFlags());
  }    
  /**
   * Construct object from file stream.
   * @param file Input stream
   * @throw IOException
   */  
  InnerClass(DataInputStream file) throws IOException
  {
    this((short)file.readUnsignedShort(), (short)file.readUnsignedShort(),
	 (short)file.readUnsignedShort(), (short)file.readUnsignedShort());
  }    
  /**
   * @param inner_class_index Class index in constant pool of inner class
   * @param outer_class_index Class index in constant pool of outer class
   * @param inner_name_index  Name index in constant pool of inner class
   * @param inner_access_flags Access flags of inner class
   */
  public InnerClass(short inner_class_index, short outer_class_index,
		    short inner_name_index, short inner_access_flags)
  {
    this.inner_class_index  = inner_class_index; 
    this.outer_class_index  = outer_class_index;
    this.inner_name_index   = inner_name_index; 
    this.inner_access_flags = inner_access_flags;
  }    
  /**
   * Called by objects that are traversing the nodes of the tree implicitely
   * defined by the contents of a Java class. I.e., the hierarchy of methods,
   * fields, attributes, etc. spawns a tree of objects.
   *
   * @param v Visitor object
   */
  public void accept(Visitor v) {
    v.visitInnerClass(this);
  }    
  /**
   * Dump inner class attribute to file stream in binary format.
   *
   * @param file Output file stream
   * @throw IOException
   */ 
  public final void dump(DataOutputStream file) throws IOException
  {
    file.writeShort(inner_class_index);
    file.writeShort(outer_class_index);
    file.writeShort(inner_name_index);
    file.writeShort(inner_access_flags);
  }    
  /**
   * @return access flags of inner class.
   */  
  public final short getInnerAccessFlags() { return inner_access_flags; }    
  /**
   * @return class index of inner class.
   */  
  public final short getInnerClassIndex() { return inner_class_index; }    
  /**
   * @return name index of inner class.
   */  
  public final short getInnerNameIndex() { return inner_name_index; }    
  /**
   * @return class index of outer class.
   */  
  public final short getOuterClassIndex() { return outer_class_index; }    
  /**
   * @param inner_access_flags.
   */
  public final void setInnerAccessFlags(short inner_access_flags) {
    this.inner_access_flags = inner_access_flags;
  }    
  /**
   * @param inner_class_index.
   */
  public final void setInnerClassIndex(short inner_class_index) {
    this.inner_class_index = inner_class_index;
  }    
  /**
   * @param inner_name_index.
   */
  public final void setInnerNameIndex(short inner_name_index) {
    this.inner_name_index = inner_name_index;
  }    
  /**
   * @param outer_class_index.
   */
  public final void setOuterClassIndex(short outer_class_index) {
    this.outer_class_index = outer_class_index;
  }    
  /**
   * @return String representation.
   */ 
  public final String toString()
  {
    return "InnerClass(" + inner_class_index + ", " + outer_class_index +
      ", " + inner_name_index + ", " + inner_access_flags;
  }    
  /**
   * @return Resolved string representation
   */ 
  public final String toString(ConstantPool constant_pool)
  { 
    String inner_class_name, outer_class_name, inner_name, access;

    inner_class_name = constant_pool.getConstantString(inner_class_index, 
						       CONSTANT_Class);
    inner_class_name = Utility.compactClassName(inner_class_name);
    outer_class_name = constant_pool.getConstantString(outer_class_index, 
						       CONSTANT_Class);
    outer_class_name = Utility.compactClassName(outer_class_name);

    inner_name = ((ConstantUtf8)constant_pool.getConstant(inner_name_index, 
							  CONSTANT_Utf8)).getBytes();
    access = Utility.accessToString(inner_access_flags, true);


    return "InnerClass:" + access + " " + inner_class_name + 
      "(\"" + outer_class_name + "\", \"" + inner_name + "\")";
  }    
}
